/*
  All purpose XML row. No moving parts, completely serviceable at home.
*/
#include <xmlrow.h>

#define TANGENT_NS_ROW "http://tangent.org/2007/06/02-row-syntax-ns#"

/* Prototype */
static void xml_row_field_st_cleanup(xml_row_st *ptr);
static void xml_row_field_st_add(xml_row_st *ptr, xmlChar *name, xmlChar *str, 
                                 size_t length, 
                                 xml_row_field_is_null is_null);
static void xml_row_release(xml_row_st *ptr);


/** @file xmlrow.c

  @brief
  XMLDB direct row functions
*/

/** @fn xml_row_create

  @brief Creates a new row object
*/
xml_row_st *xml_row_create(xml_row_st *ptr)
{
  if (!ptr)
  {
    ptr= (xml_row_st *)malloc(sizeof(xml_row_st));
    memset(ptr, 0, sizeof(xml_row_st));
    ptr->allocated= 1;
  }
  else
  {
    memset(ptr, 0, sizeof(xml_row_st));
  }

  ptr->doc= xmlNewDoc(BAD_CAST XML_DEFAULT_VERSION);

  ptr->row= xmlNewDocNode(ptr->doc, NULL, BAD_CAST "ROW", NULL);
  xmlDocSetRootElement(ptr->doc, ptr->row);
  ptr->name_space= xmlNewNs(ptr->row, 
                            (const xmlChar*)TANGENT_NS_ROW, 
                            (const xmlChar*)"row");

  return ptr;
}

xml_row_field_st *
xml_row_read(xml_row_st *ptr)
{
  xmlNodePtr field;

  assert(ptr->row);

  for (field= ptr->row->children; field; field= field->next) 
  {
    if (field->type == XML_ELEMENT_NODE)
    {
      char is_null= 0;
      xmlChar *string= NULL;
      xmlChar *name;

      if (xmlIsBlankNode(field))
        is_null= 1;
      else
        string= xmlNodeGetContent(field->children);

      name= xmlGetProp(field->children, (const xmlChar *)"name");

      xml_row_field_st_add(ptr, name, string, (size_t)xmlUTF8Strlen(string), is_null);
    }
  }

  return ptr->field;
}

xml_row_return xml_row_set(xml_row_st *ptr, char *string, size_t length)
{
  xml_row_release(ptr);

  ptr->doc= xmlParseMemory(string, length);
  if (!ptr->doc)
    return XML_ROW_FAILURE;

  ptr->row= xmlDocGetRootElement(ptr->doc);
  if (!ptr->row)
    return XML_ROW_FAILURE;

  return XML_ROW_SUCCESS;
}


char *xml_row_serialize(xml_row_st *ptr, size_t *length)
{
  xmlChar *output;
  int nlength;

  assert(ptr->row);

  if (ptr->row->children)
  {
    xmlDocDumpMemory(ptr->doc, &output, &nlength);
    *length= nlength;

    return (char *)output;
  }

  return NULL;
}

void xml_row_reset(xml_row_st *ptr)
{
  xml_row_field_st_cleanup(ptr);
}

static void xml_row_release(xml_row_st *ptr)
{
  xml_row_field_st_cleanup(ptr);
  xmlFreeDoc(ptr->doc);
}

void xml_row_close(xml_row_st *ptr)
{
  xml_row_release(ptr);
  
  if  (ptr->allocated)
    free(ptr);
}

xml_row_return xml_row_field_add(xml_row_st *ptr, 
				 char *name, 
				 char *str, 
                                 size_t length, 
                                 xml_row_field_is_null is_null)
{
  xmlNodePtr node;

  node= xmlNewChild(ptr->row, ptr->name_space, BAD_CAST "FIELD", BAD_CAST NULL);
  xmlSetProp(node, (const xmlChar *)"name", (xmlChar *)name);
  if (!is_null)
    xmlNodeSetContentLen(node, (xmlChar *)str, length);

  return XML_ROW_SUCCESS;
}


static void xml_row_field_st_add(xml_row_st *ptr, 
				 xmlChar *name, 
				 xmlChar *str, 
                                 size_t length, 
                                 xml_row_field_is_null is_null)
{
  xml_row_field_st *rptr;

  if (!ptr->field)
    ptr->fields= 0;

  if (ptr->field)
  {
    for (rptr= ptr->field; rptr && rptr->next; rptr= rptr->next); 
    rptr->next= (xml_row_field_st *)malloc(sizeof(xml_row_field_st));
    rptr= rptr->next;
  }
  else
  {
    rptr= (xml_row_field_st *)malloc(sizeof(xml_row_field_st));
    ptr->field= rptr;
  }
  memset(rptr, 0, sizeof(xml_row_field_st));

  ptr->fields++;

  rptr->name= name;
  rptr->str= str;
  rptr->length= length;
  rptr->is_null= is_null;
}

static void xml_row_field_st_cleanup(xml_row_st *ptr)
{
  xml_row_field_st *next, *fields;

  for (fields= ptr->field; fields; fields= next)
  {
    next= fields->next;

    if (fields->str)
      xmlFree(fields->str);
    if (fields->name)
      xmlFree(fields->name);

    free(fields);
  }

  ptr->field= NULL;
}

void xml_row_startup()
{
  xmlKeepBlanksDefault(0);
  xmlInitParser();
  LIBXML_TEST_VERSION;
}

void xml_row_shutdown()
{
  xmlCleanupParser();
}
